/*-----------------------------------------------------------------------------
mbgumbel.c

Shyam Prabhakar Oct 16 2003

Functions relating to the statistical significance (based on Gumbel
distribution) of high-scoring segments in a random walk.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>

#include "gumb_strucs.h"
#include "scor_strucs.h"
#include "align_strucs.h"
#include "arg_struc.h"
#include "coord_map.h"
#include "annot_ops.h"

#define MIN_BASE_RATIO 0.4     /* Min. ratio of baseline "neutral" alignment
			          columns to base-sequence positions.        */
#define MIN_BASELINE_LEN 3000  /* Min. noncoding alignment columns reqd. for
				  calculating neutral baseline statistics    */
#define L_TOL 1e-5             /* Tolerance for E[exp(lambda*s)]             */
#define YM_TOL 1e-2            /* Threshold for terminating summation of Ym  */
#define GP_SPAN 10             /* No. of steps used to estimate geometric
				  convergence ratio of Ym                    */
#define YM_EXTRAP 500          /* No. of terms in extrapolation of Ym        */


typedef struct
{

  double *p;
  int low, high;

} PROB_VEC;


void prnt_stats(PROB_VEC pv, GUMB_STATS *gs_ptr)
{

  int i, low, high;

  low  = pv.low;
  high = pv.high;

  printf("avg = %.2f, sigma = %.2f, lambda = %.8f, K = %.5f\n", gs_ptr->avg,
	 sqrt(gs_ptr->sigma2), gs_ptr->lambda, gs_ptr->K);

  printf("low = %d, high = %d\n", low, high);

  printf("prob. vector:\n");
  for (i = low; i <= high; i++)
    printf("%0.2g ", pv.p[i - low]);
  printf("\n");

}


/*-----------------------------------------------------------------------------
Return the greatest common divisor of i and j.
-----------------------------------------------------------------------------*/
int gcd(int i, int j)
{

  int tmp;

  /* make the arguments non-negative */
  if (i < 0) i = -i;
  if (j < 0) j = -j;

  /* pick the larger of the two as i */
  if (i < j)
    tmp = i, i = j, j = tmp;

  /* repeatedly replace (larger,smaller) with (smaller,larger%smaller) */
  while (j)
    tmp = i % j, i = j, j = tmp;

  /* now that smaller = 0, return larger */
  return i;

}


int get_lambda_K(PROB_VEC pv, GUMB_STATS *gs_ptr)
{

  int low, high, low_k, diff, diff_k, diff_kminus1;
  int i, j, j1, j2, k, gcd_all, gp_span;
  double *p, *tmp_p, avg, sigma2, lambda, K;
  double coef, r, lmin, lmax, lhs, delt, l_tol;
  double lhs_slope, s_k, lhs_k, *lhs_vec, Ym, conv_k_i, gp_r;
  double *p_k, *pp, *pp_k;

  p    = pv.p;
  diff = pv.high - pv.low;

  for (low = 0; low <= diff && p[low] == 0; low++)  ;
  if ( (low = pv.low + low) >= 0 )
    {
      fprintf(stderr,"ERROR: Negative scores are not possible!\n");
      return 0;
    }

  for (high = diff; high>= 0 && p[high] == 0; high--)  ;
  if ( (high = pv.low + high) <= 0 )
    {
      fprintf(stderr,"ERROR: Positive scores are not possible!\n");
      return 0;
    }

  diff = high - low;
  p    = p + low - pv.low;

  avg = low;
  for (i = 0; i <= diff; i++)
    avg += i * p[i];

  if ( avg >= 0 )
    {
      fprintf(stderr, "ERROR: Avg. score (%.2f) is >= 0!\n", avg);
      return 0;
    }

  sigma2 = -(avg - low)*(avg - low);
  for (i = 0; i <= diff; ++i)
    sigma2 += i * i * p[i];

  gcd_all = -low;
  for (i = 1; i <= diff && gcd_all != 1; i++)
    if ( p[i] )
      gcd_all = gcd(gcd_all, i + low);

  if (gcd_all > 1)
    {

      diff /= gcd_all;
      low  /= gcd_all;
      high /= gcd_all;

      tmp_p = p - gcd_all;
      p     = (double *) malloc((diff + 1)*sizeof(double));

      for (i = 0; i <= diff; p[i++] = *(tmp_p += gcd_all))  ;

    }

  lmax  = -log(p[diff])/high;  /* ubound on lambda = -log(prob(high))/high */
  lmin  = 0;
  l_tol = L_TOL/high;
  do
    {

      lambda = (lmin + lmax)/2;
      r      = exp(lambda);
      coef   = exp(lambda*low);

      lhs    = 0;
      for (i = 0; i <= diff; )
	{
	  lhs  += p[i++]*coef;
	  coef *= r;
	}

      if (lhs > 1)
	{
	  delt = lhs - 1;
	  lmax = lambda;
	}
      else
	{
	  delt = 1 - lhs;
	  lmin = lambda;
	}

    }
  while (delt > l_tol);

  coef      = exp(lambda*low);
  lhs_slope = 0;
  for (i = low; i <= high; i++)
    {
      lhs_slope += p[i-low]*i*coef;
      coef      *= r;
    }

  p_k     = (double *) calloc((int) ceil(1/YM_TOL)*diff + 1, sizeof(double));
  lhs_vec = (double *) malloc((int) (ceil(1/YM_TOL) + 1)*sizeof(double));

  *p_k = *lhs_vec = 1;
  Ym = low_k = diff_k = k = 0;
  do
    {

      low_k        += low;
      diff_kminus1  = diff_k;
      i = (diff_k += diff);
      while (i >= 0)
	{

	  if ( (j1 = i - diff_kminus1) < 0 )  j1 = 0;
	  j2 = diff < i ? diff : i;

	  pp_k = p_k + i - j1;
	  pp   = p + j1;

	  conv_k_i = 0;
	  for (j = j1; j <= j2; j++)			    
	    conv_k_i += *pp_k-- * *pp++;

	  p_k[i--] = conv_k_i;

	}

      coef  = exp(lambda*low_k);
      lhs_k = 0;
      for (i = 0; i < -low_k; i++)
	{
	  lhs_k += p_k[i]*coef;
	  coef  *= r;
	}

      while (i <= diff_k)  lhs_k += p_k[i++];

      Ym += (s_k = lhs_k/++k);
      lhs_vec[k] = lhs_k;

    }
  while (s_k > YM_TOL);

  gp_span = GP_SPAN < k ? GP_SPAN : k;
  gp_r    = pow(lhs_k/lhs_vec[k - gp_span], 1.0/gp_span);

  for (i = 0; i < YM_EXTRAP; i++)
    {
      lhs_k *= gp_r;
      Ym    += lhs_k/++k;
    }

  K = exp(-2*Ym) / (lhs_slope * (1 - 1/r));

  free(p_k);
  free(lhs_vec);
  if (gcd_all > 1)
    {
      lambda /= gcd_all;
      free(p);
    }

  gs_ptr->avg    = avg;
  gs_ptr->sigma2 = sigma2;
  gs_ptr->lambda = lambda;
  gs_ptr->K      = K;

  return 1;

}


void print_lengths(ALIGNMENT *algnmnt_ptr, int baseline_len, int col_min_seq)
{

  int nrow, i;

  nrow = algnmnt_ptr->nrow;

  for (i = 0; i < nrow; i++)
    printf("%s len = %d\n",algnmnt_ptr->seq_names[i],algnmnt_ptr->seq_lens[i]);

  printf("baseline len = %d (no. of non-exonic cols with A/C/G/T in at least "
	 "%d species)\n\n", baseline_len, col_min_seq);

}


/*-----------------------------------------------------------------------------

-----------------------------------------------------------------------------*/
void get_cms_baseline(ALIGNMENT *algnmnt_ptr, GUMB_STATS *gs_ptr,
		      ARGS *args_ptr)
{

  int i, nrow, ncol, min_row, min_col, cms, base_row;
  int *col_num_AGCT, *col_AGCT_distro, *nonexon_AGCT_distro;
  int e, n_exon, ex_start, ex_end;
  int cn, core_start, core_end, core_base_len, st, en;
  int seg, nseg, *seg_starts, *seg_ends;
  GFF_FEAT *feats;
  COORD_MAP *map_ptr;
  SEG_LIST ne_segs;

  col_AGCT_distro = algnmnt_ptr->col_AGCT_distro;
  col_num_AGCT    = algnmnt_ptr->col_num_AGCT;
  n_exon          = algnmnt_ptr->super_annot.num_feat;
  feats           = algnmnt_ptr->super_annot.feats;
  nrow            = algnmnt_ptr->nrow;
  ncol            = algnmnt_ptr->ncol;

  nonexon_AGCT_distro = (int *) malloc((nrow + 1)*sizeof(int));

  for (i = 0; i <= nrow; nonexon_AGCT_distro[i++] = *col_AGCT_distro++)  ;

  /* find alignment core (from first well-aligned position to last) and
     remove all non-core columns from nonexon_AGCT_distro */
  min_row = (nrow == 2 ? nrow : (nrow + 1)/2);
  for (i = 0; i < ncol && (cn = col_num_AGCT[i]) < min_row; i++)
    --nonexon_AGCT_distro[cn];
  core_start = i;

  for (i = ncol - 1; i > -1 && (cn = col_num_AGCT[i]) < min_row; i--)
    --nonexon_AGCT_distro[cn];
  core_end = i;

  if (core_start > core_end)
    {
      fprintf(stderr, "ERROR: couldn't find any well-aligned positions!\n");
      exit(7);
    }

  if ( (base_row = args_ptr->base) < 0 || base_row >= nrow )  base_row = 0;

  map_ptr = algnmnt_ptr->coord_maps + base_row;

  ne_segs    = annot_align_segs(algnmnt_ptr->super_annot, ncol, "nonexon");
  nseg       = ne_segs.nseg;
  seg_starts = ne_segs.starts;
  seg_ends   = ne_segs.ends;

  core_base_len = 0;
  for (seg = 0; seg < nseg; seg++)
    {
      st = seg_starts[seg] > core_start ? seg_starts[seg] : core_start;
      en = seg_ends[seg]   < core_end   ? seg_ends[seg]   : core_end;

      if ( en >= st )
	core_base_len += seq_coord(en,'<',map_ptr)-seq_coord(st,'>',map_ptr)+1;
    }

  /* remove all exonic cols in the core from nonexon_AGCT_distro */
  for (e = 0; e < n_exon; e++)
    {
      ex_start = feats[e].algn_start;
      ex_end   = feats[e].algn_end;

      if (ex_start < core_start)  ex_start = core_start;
      if (ex_end   > core_end)    ex_end   = core_end;

      for (i = ex_start; i <= ex_end; )
	--nonexon_AGCT_distro[col_num_AGCT[i++]];
    }

  /* convert to a cumulative distribution */
  for (i = nrow-1; i > -1; i--)
    nonexon_AGCT_distro[i] += nonexon_AGCT_distro[i+1];

  if ( (cms = args_ptr->col_min_seq) < 2 )
    {
      min_col = 0.5 + core_base_len * (double) MIN_BASE_RATIO;

      cms = 2;
      while (cms < nrow && nonexon_AGCT_distro[cms+1] >= min_col)  cms++;
    }
  else if (cms > nrow)
    cms = nrow;

  if ( core_base_len  < nonexon_AGCT_distro[cms] )
    {
      fprintf(stderr, "ERROR: Reference seq. \"%s\" has too few bases (%d) in "
	      "core of alignment. Pl. choose a different reference seq.\n\n",
	      algnmnt_ptr->seq_names[base_row], core_base_len);
      exit(8);
    }

  args_ptr->col_min_seq = cms;
  gs_ptr->base_ratio    = nonexon_AGCT_distro[cms]/(double) core_base_len;

/*   fprintf(stderr, "cs = %d, ce = %d, core_base_seqlen = %d, align_len = %d, cms = %d, core_baseline_len = %d, base ratio = %5.4f, baseline_ratio = %5.4f\n", core_start, core_end, core_base_len, ncol, cms, nonexon_AGCT_distro[cms], gs_ptr->base_ratio, nonexon_AGCT_distro[cms]/(double) nonexon_AGCT_distro[1]); */

}


/*-----------------------------------------------------------------------------
Return the vector of score probabilities (pv_ptr->p) required as input to
get_lambda_K().

The lowest and highest possible scores are calculated and stored in
pv_ptr->low and pv_ptr->high.

Also calculate the ratio of baseline "neutral" alignment columns to
base-sequence positions and store the value in gs_ptr->base_ratio.

Inputs:
scor_ptr points to a struct containing a vector of scores and the length of the
vector. It also contains upper and lower bounds on the scores.
Alignment columns with less than col_min_seq non-gap characters are ignored in
calculating the nonexonic score probabilities.
args_ptr contains the command-line arguments.

WARNING: the upper and lower bounds had better be valid (they don't have to be
tight, though), otherwise this function will segfault!!!
-----------------------------------------------------------------------------*/
int get_prob_vec(ALIGN_SCOR *scor_ptr, PROB_VEC *pv_ptr,
	     ALIGNMENT *algnmnt_ptr, GUMB_STATS *gs_ptr, ARGS *args_ptr)
{

  int i, min_idx, max_idx, ubound, lbound, low, high;
  int *scor_vec, len, nrow;
  double *p;
  int *counts, baseline_len, *col_num_AGCT, cms_minus1;
  int e, n_exon, ex_start;
  GFF_FEAT *feats;

  len      = scor_ptr->len;
  scor_vec = scor_ptr->scor;
  ubound   = scor_ptr->ubound;
  lbound   = scor_ptr->lbound;

  col_num_AGCT = algnmnt_ptr->col_num_AGCT;
  n_exon       = algnmnt_ptr->super_annot.num_feat;
  feats        = algnmnt_ptr->super_annot.feats;
  nrow         = algnmnt_ptr->nrow;

  if ( !len || lbound >= 0 || ubound <= 0)  return 0;

  get_cms_baseline(algnmnt_ptr, gs_ptr, args_ptr);
  cms_minus1 = args_ptr->col_min_seq - 1;

  /* the score counts vector (counts) is indexed by [score - lbound],
     so the indices go from [0] to [ubound - lbound] */
  counts = (int *) calloc(ubound - lbound + 1, sizeof(int));

  /* columns containing less than col_min_seq nucleotide characters are not
     considered for calculating score probabilities (neutral baseline) */
  i = -1, baseline_len = 0;
  for (e = 0; e < n_exon; e++)
    {
      ex_start = feats[e].algn_start;

      while ( ++i < ex_start )
	if ( col_num_AGCT[i] > cms_minus1 )
	  {
	    ++counts[ scor_vec[i] - lbound ];
	    ++baseline_len;
	  }

      i = feats[e].algn_end;
    }

  while ( ++i < len )
    if ( col_num_AGCT[i] > cms_minus1 )
      {
	++counts[ scor_vec[i] - lbound ];
	++baseline_len;
      }

  if ( baseline_len < MIN_BASELINE_LEN )
    {
      fprintf(stderr, "ERROR: only %d noncoding alignment columns available"
	      " for calculating neutral baseline statistics, require at"
	      " least %d!\n", baseline_len, MIN_BASELINE_LEN);
      exit(5);
/*       return 0;  */
    }

  min_idx = -1;
  while ( !counts[++min_idx] )  ;
  low = min_idx + lbound;

  max_idx = ubound - lbound + 1;
  while ( !counts[--max_idx] )  ;
  high = max_idx + lbound;

  /* lowest score must be negative, and highest score must be positive */
  if (low >= 0 || high <= 0)
    {
      free(counts);
      return 0;
    }

  p = (double *) calloc(high - low + 1, sizeof(double));

  for (i = min_idx; i <= max_idx; i++)
    p[i + lbound - low] = counts[i]/ (double) baseline_len;

  print_lengths(algnmnt_ptr, baseline_len, args_ptr->col_min_seq);

  free(counts);

  pv_ptr->p    = p;
  pv_ptr->low  = low;
  pv_ptr->high = high;

  return 1;

}


/*-----------------------------------------------------------------------------
Calculate the lambda and K parameters of the Gumbel distribution.

Inputs:
scor_ptr points to a struct containing a vector of scores and the length of the
vector. It also contains upper and lower bounds on the scores.
col_num_AGCT is an array of length ncol (alignment length) containing the
number of aligned nucleotides in each column (excluding 'N's).
args_ptr contains the command-line arguments.

Output:
lambda, K, and the mean and variance of the letter scores are stored in the
struct pointed to by gs_ptr, as well as the ratio of baseline "neutral"
alignment columns to base-sequence positions.
-----------------------------------------------------------------------------*/
int mbgumbel(ALIGN_SCOR *scor_ptr, ALIGNMENT *algnmnt_ptr, GUMB_STATS *gs_ptr,
	     ARGS *args_ptr)
{

  PROB_VEC pv;

  if ( !get_prob_vec(scor_ptr, &pv, algnmnt_ptr, gs_ptr, args_ptr) ||
       !get_lambda_K(pv, gs_ptr) )
    {
      fprintf(stderr, "mbgumbel(): couldn't calculate lambda, K!\n");
      return 0;  /* failure */
    }

/*   prnt_stats(pv, gs_ptr); */
/*   exit(1); */

  free(pv.p);
  return 1;  /* success */

}


/*-----------------------------------------------------------------------------
Returns the score corresponding to a p-value p, given lambda, K and M.
-----------------------------------------------------------------------------*/
float pval2scor(GUMB_STATS gs, double p, int M)
{

  double lambda, K;
  float scor;

  if (p >= 1 || p <= 0)  return 0;

  lambda = gs.lambda;
  K      = gs.K;

  /* if there's no possibility of divide-by-zero error, return the score */
  if (lambda && K && M)
    if (p > 1e-6)  scor = log(-(K*M) / log(1-p)) /lambda;
    else           scor = log( K*M/p ) /lambda;
  else
    {
      fprintf(stderr,"ERROR: pval2scor(): cannot calculate threshold score\n");
      exit(6);
    }

  return scor >= 0 ? scor : 0;
    
}


/*-----------------------------------------------------------------------------
Returns the p-value of a score x, given lambda, K and M.
If the p-value is less than 0.1 (approx.), then a Taylor expansion is used.
-----------------------------------------------------------------------------*/
double pvalue(double l, double k, int m, double x)
{

  double y = k * m * exp(-l*x);

  return y > 0.1 ? 1 - exp(-y) : y - y*y/2;

}
