/*-----------------------------------------------------------------------------
parsim.c

Shyam Prabhakar Aug 21 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "align_ops.h"
#include "scor_strucs.h"
#include "arg_struc.h"
#include "classify_annots.h"
#include "annot_ops.h"

#define SCOR_SCALE 200
#define SEG_ALLOC_SIZ 30


int calc_PLL_scor(ALIGNMENT *tree_algn_ptr, float **cons_scor, float mut_scor,
		  float **N_scor, float **mut_freq_mat,
		  ALIGN_SCOR *col_scor_ptr, ALIGN_SCOR *stat_scor_ptr)
{

  int nrow, ncol, row, col, nmatch;
  unsigned char **matrix;
  int *col_num_AGCT, *c_scor_vec, *s_scor_vec;
  float align_scor, gap_scor;
  int is_gap[256] = {0};
  char c1, c2;

  if ( !tree_algn_ptr || (nrow = tree_algn_ptr->nrow) < 2 ||
       !(ncol = tree_algn_ptr->ncol) )
    {
      fprintf(stderr, "parsim: invalid input alignment!\n");
      return 0;
    }

  col_num_AGCT = tree_algn_ptr->col_num_AGCT;

  col_scor_ptr->len  = stat_scor_ptr->len = ncol;
  col_scor_ptr->scor  = c_scor_vec = (int *) calloc(ncol, sizeof(int));
  stat_scor_ptr->scor = s_scor_vec = (int *) calloc(ncol, sizeof(int));

  nmatch = nrow > 2 ? nrow : 1;
  matrix = (unsigned char **) calloc(nmatch + 1, sizeof(unsigned char *));

  /* It is assumed that the rows in the alignment matrix are already ordered
     to form a circular tour of the leaves of the phylogenetic tree. */
  for (row = 0; row < nrow; row++)
    matrix[row] = tree_algn_ptr->matrix[row];

  if (nrow > 2)  matrix[nrow] = tree_algn_ptr->matrix[0];

  identify_gap_chars(is_gap);

  for (col = 0; col < ncol; col++)
    if ( *col_num_AGCT++ == 0 )  c_scor_vec++, s_scor_vec++;
    else
      {
	
	align_scor = gap_scor = row = 0;
	c2   = matrix[0][col];

	/* NOTE gap<->gap is not penalised; only gap<->nongap is penalised */
	while ( row++ < nmatch && (c1 = c2) )
	  if ( is_gap[ c2 = matrix[row][col] ] != is_gap[ c1 ] )
	    gap_scor += N_scor[row-1][row];
	  else if ( !is_gap[ c1 ] && c1 != 'N' && c2 != 'N' )
	    align_scor += c1 == c2 ? cons_scor[row-1][row] : mut_scor;

	*c_scor_vec++ = (int) floor(0.5 + align_scor + gap_scor);
	*s_scor_vec++ = (int) floor(0.5 + align_scor);

      }

  free(matrix);

  return 1;

}


void print_scor_mat(float **scor_mat, int nrow, char **seq_names)
{

  int i, j;

  printf("     ");
  for (j = 0; j < nrow; j++)  printf("  %-5.5s", seq_names[j]);
  printf("\n");

  for (i = 0; i < nrow; i++)
    {
      printf("%-5.5s", seq_names[i]);

      for (j = 0; j < nrow; j++)  printf(" %5.1f ", scor_mat[i][j]);
      printf("\n");
    }

}


float **log_like_N_scor(float **mut_freq_mat, int nrow, float ratio)
{

  int i, j;
  float mut_freq, mut_scor, cons_scor, **N_scor_mat;

  N_scor_mat  = (float **) malloc(nrow * sizeof(float *));
  for (i = 0; i < nrow; i++)
    N_scor_mat[i]  = (float *) calloc(nrow + 1, sizeof(float));

  mut_scor = -log(ratio) * SCOR_SCALE;
  for (i = 0; i < nrow; i++)
    for (j = i+1; j < nrow; j++)
      {

	mut_freq  = mut_freq_mat[i][j];
	cons_scor = SCOR_SCALE * log((1 - mut_freq/ratio)/(1 - mut_freq));

	N_scor_mat[i][j] = N_scor_mat[j][i] =
	  mut_freq*mut_scor + (1 - mut_freq)*cons_scor;

      }

  /* NOTE: hack to allow circular tour scoring */
  for (i = 0; i < nrow; i++)
    N_scor_mat[i][nrow] = N_scor_mat[i][0];

  return N_scor_mat;

}


float **log_like_cons_scor(float **mut_freq_mat, int nrow, float ratio)
{

  int i, j;
  float mut_freq, **cons_scor_mat;

  cons_scor_mat  = (float **) malloc(nrow * sizeof(float *));
  for (i = 0; i < nrow; i++)
    cons_scor_mat[i]  = (float *) calloc(nrow + 1, sizeof(float));

  for (i = 0; i < nrow; i++)
    for (j = i+1; j < nrow; j++)
      {
	mut_freq = mut_freq_mat[i][j];
	cons_scor_mat[i][j] = cons_scor_mat[j][i] =
	  SCOR_SCALE * log((1 - mut_freq/ratio)/(1 - mut_freq));
      }

  /* NOTE: hack to allow circular tour scoring */
  for (i = 0; i < nrow; i++)
    cons_scor_mat[i][nrow] = cons_scor_mat[i][0];

  return cons_scor_mat;

}


void print_mut_freqs(float **freq_mat, int nrow, char **seq_names)
{

  int i, j;

  fprintf(stdout, "     ");
  for (j = 0; j < nrow; j++)  fprintf(stdout, "  %-5.5s", seq_names[j]);
  fprintf(stdout, "\n");

  for (i = 0; i < nrow; i++)
    {
      fprintf(stdout, "%-5.5s", seq_names[i]);

      for (j = 0; j < nrow; j++)  fprintf(stdout, "  %.3f", freq_mat[i][j]);
      fprintf(stdout, "\n");
    }

  fprintf(stdout, "\n");

}


float mut_freq(char *seq1, char *seq2, SEG_LIST seg_list1, SEG_LIST seg_list2,
	       int min_blk_siz)
{

  int i, nmut, naligned, is_ATGC[256] = {0};
  int s1, s2, nseg1, nseg2;
  int *starts1, *starts2, *ends1, *ends2;
  int start1, start2, end1, end2,  max_start, min_end;
  unsigned char n1, n2;

  nseg1   = seg_list1.nseg;
  starts1 = seg_list1.starts;
  ends1   = seg_list1.ends;

  nseg2   = seg_list2.nseg;
  starts2 = seg_list2.starts;
  ends2   = seg_list2.ends;

  is_ATGC[(unsigned char) 'A'] = is_ATGC[(unsigned char) 'T'] =
  is_ATGC[(unsigned char) 'G'] = is_ATGC[(unsigned char) 'C'] = 1;

  if ( min_blk_siz < 1)  min_blk_siz = 1;

  nmut = naligned = 0;
  s2   = 0;
  for (s1 = 0; s1 < nseg1; s1++)
    {

      start1 = starts1[s1];
      end1   = ends1[s1];

      while (s2 < nseg2)
	if ( (end2 = ends2[s2]) >=  start1 )
	  {  /* seg2 either intersects seg1 or is to the right of seg1 */

	    start2 = starts2[s2];

	    max_start = start2 > start1 ? start2 : start1;
	    min_end   = end2   < end1   ? end2   : end1;

	    /* if there's an intersection of length >= min_blk_siz */
	    if ( min_end - max_start > min_blk_siz - 2 )
	      for (i = max_start; i <= min_end; i++)  /* count mutations */
		if ( is_ATGC[ n1 = seq1[i] ] && is_ATGC[ n2 = seq2[i] ] )
		  {
		    naligned++;
		    if (n1 != n2)  nmut++;
		  }

	    if (end2 >= end1)  break;
	    else               s2++;

	  }
	else  s2++;

    }

  if (naligned)  return nmut/ (float) naligned;
  else           return -1.0;  /* error */

}


SEG_LIST get_NC_seg_list(COORD_MAP map, GFF_ANNOT annot)
{

  int m, f, array_siz, num_feat, feat_start, feat_end;
  int map_siz, *algn_coords, *seq_coords;
  int nseg, *starts, *ends, seg_start, seg_end;
  char *feat_name;
  SEG_LIST seg_list = {0};
  GFF_FEAT *feats;

  map_siz     = map.map_siz;
  algn_coords = map.algn_coords;
  seq_coords  = map.seq_coords;

  num_feat = annot.num_feat;
  feats    = annot.feats;

  starts = (int *) malloc(0);
  ends   = (int *) malloc(0);
  nseg = f = 0;

  /* the first map is -1 -> -1, and the last is align_len -> seq_len */
  for (m = 1; m < map_siz - 1; m++)
    {

      seg_start = algn_coords[m];
      seg_end   = seg_start + seq_coords[m + 1] - seq_coords[m] - 1;

      while ( f < num_feat && (feat_name = feats[f].feature) )
	if ( !is_exon_cds_utr(feat_name) ||
	     (feat_end = feats[f].algn_end) < seg_start )
	  {  /* either not an exon, or exon to the left of the segment */
	    f++;
	  }
	else if ( feat_end < seg_end )
	  {  /* exon overlaps ungapped segment, but is not
		necessarily the rightmost exon that does so */

	    if ( (feat_start = feats[f].algn_start) > seg_start )
	      {
		/* exon breaks the segment in two ...
		   save the left segment and move on */
		if ( !(nseg % SEG_ALLOC_SIZ) )
		  {
		    array_siz = nseg + SEG_ALLOC_SIZ;

		    starts = (int *) realloc(starts, array_siz * sizeof(int));
		    ends   = (int *) realloc(ends,   array_siz * sizeof(int));
		  }

		starts[nseg] = seg_start;
		ends[nseg++] = feat_start - 1;
	      }

	    seg_start = feat_end + 1;
	    f++;

	  }
	else  /* rightmost exon with a chance of overlapping the segment */
	  {
	    /* if exon overlaps segment, modify seg_end */
	    if ( (feat_start = feats[f].algn_start) <= seg_end )
	      seg_end = feat_start - 1;

	    break;
	  }

      /* seg_end will be < seg_start if exons completely cover the seg. */
      if (seg_end >= seg_start)
	{
	  if ( !(nseg % SEG_ALLOC_SIZ) )
	    {
	      array_siz = nseg + SEG_ALLOC_SIZ;

	      starts = (int *) realloc(starts, array_siz * sizeof(int));
	      ends   = (int *) realloc(ends,   array_siz * sizeof(int));
	    }

	  starts[nseg] = seg_start;
	  ends[nseg++] = seg_end;
	}

    }

  seg_list.nseg   = nseg;
  seg_list.starts = starts;
  seg_list.ends   = ends;

  return seg_list;

}


float **mut_freqs(ALIGNMENT *algnmnt_ptr, int min_blk_len)
{

  int i, j, nrow;
  float mf, **freq_mat;
  char **mat, **seq_names;
  COORD_MAP *maps;
  SEG_LIST *seg_lists;

  nrow      = algnmnt_ptr->nrow;
  mat       = algnmnt_ptr->matrix;
  seq_names = algnmnt_ptr->seq_names;
  maps      = algnmnt_ptr->coord_maps;

  seg_lists = (SEG_LIST *) malloc(nrow * sizeof(SEG_LIST));
  for (i = 0; i < nrow; i++)
    seg_lists[i] = get_NC_seg_list(maps[i], algnmnt_ptr->super_annot);

  freq_mat  = (float **) malloc(nrow * sizeof(float *));
  for (i = 0; i < nrow; i++)
    freq_mat[i]  = (float *) calloc(nrow, sizeof(float));

  for (i = 0; i < nrow; i++)
    for (j = i+1; j < nrow; j++)
      if ( (mf = mut_freq(mat[i], mat[j], seg_lists[i],
			  seg_lists[j], min_blk_len)) < 0 )
	{
	  freq_mat[i][j] = freq_mat[j][i] = 0.0;
	  fprintf(stderr, "WARNING: %s and %s have no aligned noncoding bases"
		  " in blocks of size >= %d!\n", seq_names[i], seq_names[j],
		  min_blk_len);
	}
      else
	freq_mat[i][j] = freq_mat[j][i] = mf;

/*   printf("%.3f\n", freq_mat[0][1]); */
  print_mut_freqs(freq_mat, nrow, seq_names);

  return freq_mat;

}


int parsim_log_like(ALIGNMENT *algnmnt_ptr, ARGS args,
		    ALIGN_SCOR *col_scor_ptr, ALIGN_SCOR *stat_scor_ptr)
{

  int nrow, i, j;
  float **mut_freq_mat;
  float mut_scor, **cons_scor_mat, **N_scor_mat;
  int ubound, blk_len;

/*   print_annots(transfer_annot(algnmnt_ptr->annots[0], algnmnt_ptr->coord_maps + 3, algnmnt_ptr->seq_names[3])); */

  blk_len      = args.blk_len;
  mut_freq_mat = mut_freqs(algnmnt_ptr, blk_len);

  nrow          = algnmnt_ptr->nrow;
  cons_scor_mat = log_like_cons_scor(mut_freq_mat, nrow, args.mut_ratio);

  mut_scor     = -log(args.mut_ratio) * SCOR_SCALE;
  N_scor_mat   = log_like_N_scor(mut_freq_mat, nrow, args.mut_ratio);

/*   printf("mut_scor = %.1f\n", mut_scor); */
/*   print_scor_mat(cons_scor_mat, nrow, algnmnt_ptr->seq_names); */
/*   print_scor_mat(N_scor_mat, nrow, algnmnt_ptr->seq_names); */
/*   exit(1); */

  if ( !calc_PLL_scor(algnmnt_ptr, cons_scor_mat, mut_scor, N_scor_mat,
		      mut_freq_mat, col_scor_ptr, stat_scor_ptr) )
    return 0;

/*   for (j = 0; j < algnmnt_ptr->ncol; j++) */
/*     { */
/*       for (i = 0; i < algnmnt_ptr->nrow; i++) */
/* 	printf(" %c", algnmnt_ptr->matrix[i][j]); */

/*       printf("; scor = %d ", PLL_scor_ptr->scor[j]); */
/*       printf("\n"); */
/*     } */
/*   exit(1); */

  ubound = 0;
  for (i = 0; i < nrow; i++)
    ubound += ceil( cons_scor_mat[i][ (i+1)%nrow ] );

  col_scor_ptr->lbound = stat_scor_ptr->lbound = floor( mut_scor*nrow );
  col_scor_ptr->ubound = stat_scor_ptr->ubound = ubound;

  return 1;

}
